import pytest

import gmpy2
from gmpy2 import const_catalan, const_euler, const_log2, const_pi, mpfr


def test_pi():
    assert const_pi() == mpfr('3.1415926535897931')
    assert const_pi(100) == mpfr('3.1415926535897932384626433832793',100)

    pytest.raises(TypeError, lambda: const_pi(100,200))
    pytest.raises(TypeError, lambda: const_pi(prec=100))

    assert const_pi(precision=100) == mpfr('3.1415926535897932384626433832793',100)
    assert gmpy2.ieee(32).const_pi() == mpfr('3.14159274',24)

    pytest.raises(TypeError, lambda: gmpy2.ieee(32).const_pi(100))

    assert gmpy2.ieee(128).const_pi() == mpfr('3.1415926535897932384626433832795028',113)


def test_log2():
    assert const_log2() == mpfr('0.69314718055994529')
    assert const_log2(100) == mpfr('0.69314718055994530941723212145798',100)

    pytest.raises(TypeError, lambda: const_log2(100,200))
    pytest.raises(TypeError, lambda: const_log2(prec=100))

    assert const_log2(precision=100) == mpfr('0.69314718055994530941723212145798',100)
    assert gmpy2.ieee(32).const_log2() == mpfr('0.693147182',24)

    pytest.raises(TypeError, lambda: gmpy2.ieee(32).const_log2(100))

    assert gmpy2.ieee(128).const_log2() == mpfr('0.693147180559945309417232121458176575',113)


def test_catalan():
    assert const_catalan() == mpfr('0.91596559417721901')
    assert const_catalan(100) == mpfr('0.91596559417721901505460351493252',100)

    pytest.raises(TypeError, lambda: const_catalan(100,200))
    pytest.raises(TypeError, lambda: const_catalan(prec=100))

    assert gmpy2.const_catalan(precision=100) == mpfr('0.91596559417721901505460351493252',100)
    assert gmpy2.ieee(32).const_catalan() == mpfr('0.915965617',24)

    pytest.raises(TypeError, lambda: gmpy2.ieee(32).const_catalan(100))

    assert gmpy2.ieee(128).const_catalan() == mpfr('0.915965594177219015054603514932384146',113)


def test_euler():
    assert const_euler() == mpfr('0.57721566490153287')
    assert const_euler(100) == mpfr('0.57721566490153286060651209008234',100)

    pytest.raises(TypeError, lambda: const_euler(100,200))
    pytest.raises(TypeError, lambda: const_euler(prec=100))

    assert const_euler(precision=100) == mpfr('0.57721566490153286060651209008234',100)
    assert gmpy2.ieee(32).const_euler() == mpfr('0.577215672',24)

    pytest.raises(TypeError, lambda: gmpy2.ieee(32).const_euler(100))

    assert gmpy2.ieee(128).const_euler() == mpfr('0.577215664901532860606512090082402471',113)



from ctypes import memmove

import pytest
from hypothesis import example, given, settings
from hypothesis.strategies import integers

import gmpy2
from gmpy2 import from_binary, mpfr, mpq, mpz, to_binary, xmpz


def test_xmpz_digits():
    x = xmpz(6)
    assert x.digits() == '6'
    assert x.digits(2) == '0b110'
    assert x.digits(8) == '0o6'
    assert x.digits(16) == '0x6'
    assert xmpz(30).digits(16) == '0x1e'

    pytest.raises(ValueError, lambda: x.digits(0))
    pytest.raises(ValueError, lambda: x.digits(63))


def test_xmpz_limbs():
    x, y = xmpz(123456789), xmpz(0);
    x_limbs, num_limbs = x.limbs_read(), x.num_limbs()

    assert x_limbs != 0
    assert num_limbs > 0

    y_limbs = y.limbs_write(num_limbs)

    assert y_limbs != 0
    assert memmove(y_limbs, x_limbs, num_limbs * xmpz.limb_size) > 0

    y.limbs_finish(num_limbs)

    assert int(y) == 123456789

    x, y = xmpz(987654321), xmpz(0)
    x_limbs, num_limbs = x.limbs_read(), x.num_limbs()

    assert x_limbs != 0
    assert num_limbs > 0

    y_limbs = y.limbs_modify(num_limbs);

    assert y_limbs != 0
    assert memmove(y_limbs, x_limbs, num_limbs * xmpz.limb_size) > 0

    y.limbs_finish(num_limbs)

    assert int(y) == 987654321


def test_xmpz_attributes():
    x = xmpz(10)

    assert x.numerator == xmpz(10)
    assert x.denominator == xmpz(1)
    assert x.real == xmpz(10)


def test_xmpz_misc():
    z = xmpz(5)

    assert gmpy2.xbit_mask(7) == xmpz(127)
    assert gmpy2.xbit_mask(z) == xmpz(31)

    pytest.raises(TypeError, lambda: gmpy2.xbit_mask(4.5))

    assert not xmpz(0)
    assert z
    -z
    assert z == xmpz(-5)
    -z
    assert z == xmpz(5)
    +z
    assert z == xmpz(5)

    z2 = z.copy()

    assert z2 == z
    assert z2 is not z
    assert z.copy().make_mpz() == mpz(5)
    assert xmpz(-100).make_mpz() == mpz(-100)
    assert len(z) == 3
    assert len(xmpz(3000)) == 12

    assert str(xmpz(42)) == '42'
    assert repr(xmpz(42)) == 'xmpz(42)'


def test_xmpz_create():
    pytest.raises(TypeError, lambda: xmpz(s=1))
    pytest.raises(TypeError, lambda: xmpz(1, s=2))


def test_xmpz_subscripts():
    x = xmpz(10)

    assert x[0] == x[2] == 0
    assert x[1] == x[3] == 1

    x[0] = 1

    assert x == xmpz(11)
    assert x[0:2] == mpz(3)
    assert x[0:] == mpz(11)
    assert x[:4] == mpz(11)

    x[:4] = 14

    assert x == xmpz(14)
    assert x[0] == 0

    pytest.raises(TypeError, lambda: x[mpfr('inf')])

    assert x[3:0] == mpz(0)
    assert x[-1] == 1

    x[-1] = 0

    assert x == xmpz(6)

    x[0:4] = 15

    assert x == xmpz(15)

    with pytest.raises(ValueError):
        x[0] = 16

    x[0:4] = 16

    assert x == xmpz(0)

    x[0:4] = 15

    assert x == xmpz(15)

    x[0:5] = 16

    assert x == xmpz(16)


def test_xmpz_iterators():
    x = xmpz(16)

    assert [b for b in x.iter_bits()] == [False, False, False, False, True]

    x = xmpz(30)

    assert [b for b in x.iter_bits()] == [False, True, True, True, True]
    assert [b for b in x.iter_bits(-1, 2)] == []
    assert [b for b in x.iter_bits(1, 3)] == [True, True]
    assert [b for b in x.iter_set()] == [1, 2, 3, 4]
    assert [b for b in x.iter_clear()] == [0]

    x = xmpz(10)

    assert [b for b in x.iter_clear()] == [0, 2]


def test_xmpz_conversion():
    assert xmpz('5') == xmpz(5)
    assert xmpz('5') == xmpz(5)

    pytest.raises(ValueError, lambda: xmpz('not'))

    assert xmpz(-3.5) == xmpz(-3)

    pytest.raises(OverflowError, lambda: xmpz(float('inf')))

    assert xmpz(mpz(100)) == xmpz(100)
    assert xmpz(xmpz(100)) == xmpz(100)
    assert xmpz(mpq(30,2)) == xmpz(15)
    assert str(xmpz(100)) == '100'


def test_xmpz_abs():
    a = xmpz(123)
    b = abs(a)

    assert a is not b

    a = xmpz(-123)
    b = abs(a)

    assert a == xmpz(123)
    assert b is None

    b = abs(a)

    assert b is None


def test_xmpz_iadd():
    x = xmpz(5)
    x += mpz(6)

    assert x == xmpz(11)

    x += 7

    assert x == xmpz(18)

    x += -6

    assert x == xmpz(12)

    x += mpfr(2.5)

    assert x == mpfr('14.5')


def test_xmpz_isub():
    x = xmpz(7)
    x -= xmpz(1)

    assert x == xmpz(6)

    x -= 1

    assert x == xmpz(5)

    x -= mpz(7)

    assert x == xmpz(-2)

    x -= -5

    assert x == xmpz(3)

    x -= -mpfr(5)

    assert x == mpfr('8.0')


def test_xmpz_imul():
    x = xmpz(2)
    x *= xmpz(2)

    assert x == xmpz(4)

    x *= 2

    assert x == xmpz(8)

    x *= mpz(3)

    assert x == xmpz(24)

    x *= -1

    assert x == xmpz(-24)

    x *= mpfr(-0.5)

    assert x == mpfr('12.0')

def test_xmpz_ifloordiv():
    x = xmpz(49)
    x //= xmpz(3)

    assert x == xmpz(16)

    x //= mpz(3)

    assert x == xmpz(5)

    x //= 2

    assert x == xmpz(2)

    with pytest.raises(ZeroDivisionError):
        x //= 0

    x == xmpz(2)
    x //= mpfr(-0.5)

    assert x == mpfr('-4.0')


def test_xmpz_imod():
    x = xmpz(45)
    x %= xmpz(18)

    assert x == xmpz(9)

    x %= mpz(2)

    assert x == xmpz(1)

    x = xmpz(40)
    x %= 21

    assert x == xmpz(19)

    with pytest.raises(ZeroDivisionError):
        x %= 0

    x == xmpz(19)
    x %= mpfr(10)

    assert x == mpfr('9.0')


def test_xmpz_ishifts():
    x = xmpz(63)
    x >>= xmpz(63)

    assert x == xmpz(0)

    x = xmpz(63)
    x >>= xmpz(1)

    assert x == xmpz(31)

    x >>= mpz(2)

    assert x == xmpz(7)

    x >>= 1

    assert x == xmpz(3)

    x <<= xmpz(2)

    assert x == xmpz(12)

    x <<= mpz(1)

    assert x == xmpz(24)

    x <<= 0

    assert x == xmpz(24)

    with pytest.raises(TypeError):
        x >>= mpfr(2)
    with pytest.raises(TypeError):
        x <<= mpfr(2)
    with pytest.raises(OverflowError):
        x >>= -1
    with pytest.raises(OverflowError):
        x <<= -5


def test_xmpz_ipow():
    x = xmpz(5)
    x **= xmpz(2)

    assert x == xmpz(25)

    x **= mpz(2)

    assert x == xmpz(625)

    with pytest.raises(OverflowError):
        x **= -2

    x **= 2

    assert x == xmpz(390625)

    with pytest.raises(TypeError):
        x **= mpfr(2)


def test_xmpz_iand():
    x = xmpz(7)
    x &= xmpz(5)

    assert x == xmpz(5)

    x &= mpz(4)

    assert x == xmpz(4)

    x &= 9

    assert x == xmpz(0)

    x = mpz(4)
    x &= 12

    assert x == mpz(4)

    with pytest.raises(TypeError):
        x &= mpfr(4)


def test_xmpz_ior():
    x = xmpz(0)
    x |= xmpz(1)

    assert x == xmpz(1)

    x |= xmpz(0)

    assert x == xmpz(1)

    x = xmpz(0)
    x |= xmpz(0)

    assert x == xmpz(0)

    x |= 5

    assert x == xmpz(5)

    with pytest.raises(TypeError):
        x |= mpfr(3)

def test_xmpz_ixor():
    x = xmpz(1)
    x ^= xmpz(0)

    assert x == xmpz(1)

    x ^= mpz(1)

    assert x == xmpz(0)

    x ^= 1

    assert x == xmpz(1)

    with pytest.raises(TypeError):
        x ^= mpfr(0)


@settings(max_examples=1000)
@example(0)
@example(1)
@example(-1)
@example(1234567890123456789)
@given(integers())
def test_xmpz_to_from_binary(x):
    x = xmpz(x)

    assert from_binary(to_binary(x)) == x



import os
import platform
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from pathlib import Path

ON_WINDOWS = platform.system() == 'Windows'
_comp_args = ["DSHARED=1"]
sources = ['src/gmpy2.c']
if os.getenv('CIBUILDWHEEL'):
    include_dirs = [os.path.join(os.path.dirname(__file__), '.local', 'include')]
    library_dirs = [os.path.join(os.path.dirname(__file__), '.local',
                                 'bin' if ON_WINDOWS else 'lib')]
else:
    include_dirs = []
    library_dirs = []

class Gmpy2Build(build_ext):
    description = "Build gmpy2 with custom build options"
    user_options = build_ext.user_options + [
        ('fast', None,
         "Depend on MPFR and MPC internal implementations details"
         "(even more than the standard build)"),
        ('gcov', None, "Enable GCC code coverage collection"),
        ('vector', None, "Include the vector_XXX() functions;"
         "they are unstable and under active development"),
        ('static', None, "Enable static linking compile time options."),
        ('static-dir=', None, "Enable static linking and specify location."),
        ('gdb', None, "Build with debug symbols."),
    ]

    def initialize_options(self):
        build_ext.initialize_options(self)
        self.fast = False
        self.gcov = False
        self.vector = False
        self.static = False
        self.static_dir = False
        self.gdb = False

    def finalize_options(self):
        build_ext.finalize_options(self)
        self.force = 1
        if self.fast:
            _comp_args.append('DFAST=1')
        if self.gcov:
            if ON_WINDOWS:
                raise ValueError("Cannot enable GCC code coverage on Windows")
            _comp_args.append('DGCOV=1')
            _comp_args.append('O0')
            _comp_args.append('-coverage')
            self.libraries.append('gcov')
        if self.vector:
            _comp_args.append('DVECTOR=1')
        if self.static:
            _comp_args.remove('DSHARED=1')
            _comp_args.append('DSTATIC=1')
        if self.gdb:
            _comp_args.append('ggdb')
        if self.static_dir:
            _comp_args.remove('DSHARED=1')
            _comp_args.append('DSTATIC=1')
            self.include_dirs.append(self.static_dir + '/include')
            self.library_dirs.append(self.static_dir + '/lib')

    def build_extensions(self):
        compiler = self.compiler.compiler_type
        _prefix = '-' if compiler != 'msvc' else '/'
        for i in range(len(_comp_args)):
            _comp_args[i] = ''.join([_prefix, _comp_args[i]])
        build_ext.build_extensions(self)

extensions = [
    Extension('gmpy2.gmpy2',
              sources=sources,
              include_dirs=include_dirs,
              libraries=['mpc','mpfr','gmp'] + ([] if ON_WINDOWS else ['m']),
              library_dirs=library_dirs,
              extra_compile_args=_comp_args,
              )
]

cmdclass = {'build_ext': Gmpy2Build}

setup(
    cmdclass=cmdclass,
    ext_modules=extensions,
)
